import logging
from pandas import DataFrame

from autogluon.core.constants import BINARY, MULTICLASS, REGRESSION, QUANTILE

logger = logging.getLogger(__name__)


# Cleaner cleans data prior to entering feature generation
class Cleaner:
    @staticmethod
    def construct(problem_type: str, label: str, threshold: int):
        if problem_type == BINARY:
            return CleanerDummy()
        elif problem_type == MULTICLASS:
            return CleanerMulticlass(label=label, threshold=threshold)
        elif problem_type in [REGRESSION, QUANTILE]:
            return CleanerDummy()
        else:
            raise NotImplementedError

    def fit(self, X: DataFrame) -> DataFrame:
        raise NotImplementedError

    def fit_transform(self, X: DataFrame) -> DataFrame:
        self.fit(X)
        return self.transform(X)

    def transform(self, X: DataFrame) -> DataFrame:
        raise NotImplementedError


class CleanerDummy(Cleaner):
    def __init__(self):
        pass

    def fit(self, X: DataFrame) -> DataFrame:
        pass

    def transform(self, X: DataFrame) -> DataFrame:
        return X


class CleanerMulticlass(Cleaner):
    def __init__(self, label: str, threshold: int):
        self.label = label
        self.threshold = threshold
        self.valid_classes = None

    def fit(self, X: DataFrame):
        self.valid_classes = self.get_valid_classes(X=X, label=self.label, threshold=self.threshold)

    def transform(self, X: DataFrame) -> DataFrame:
        return self.remove_classes(X=X, label=self.label, valid_classes=self.valid_classes)

    @staticmethod
    def get_valid_classes(X, label, threshold):
        class_counts = X[label].value_counts()
        class_counts_valid = class_counts[class_counts >= threshold]
        valid_classes = list(class_counts_valid.index)
        sum_prior = sum(class_counts)
        sum_after = sum(class_counts_valid)
        percent = sum_after / sum_prior
        if len(valid_classes) < len(class_counts):
            logger.log(25, 'Warning: Some classes in the training set have fewer than %s examples. AutoGluon will only keep %s out of %s classes for training and will not try to predict the rare classes. '
                           'To keep more classes, increase the number of datapoints from these rare classes in the training data or reduce label_count_threshold.' % (threshold, len(valid_classes), len(class_counts)))
        if percent < 1.0:
            logger.log(25, 'Fraction of data from classes with at least %s examples that will be kept for training models: %s' % (threshold, percent))
        return valid_classes

    @staticmethod
    def remove_classes(X, label, valid_classes):
        X = X[X[label].isin(valid_classes)]
        return X



import autograd.numpy as anp
import autograd.scipy.linalg as aspl
from autograd.builtins import isinstance
from autograd.tracer import getval

from .constants import \
    NOISE_VARIANCE_LOWER_BOUND, MIN_POSTERIOR_VARIANCE, \
    MIN_CHOLESKY_DIAGONAL_VALUE
from .custom_op import AddJitterOp, cholesky_factorization, flatten_and_concat


def cholesky_computations(
        features, targets, mean, kernel, noise_variance, debug_log=False,
        test_intermediates=None):
    """
    Given input matrix X (features), target matrix Y (targets), mean and kernel
    function, compute posterior state {L, P}, where L is the Cholesky factor
    of
        k(X, X) + sigsq_final * I
    and
        L P = Y - mean(X)
    Here, sigsq_final >= noise_variance is minimal such that the Cholesky
    factorization does not fail.

    :param features: Input matrix X (n,d)
    :param targets: Target matrix Y (n,m)
    :param mean: Mean function
    :param kernel: Kernel function
    :param noise_variance: Noise variance (may be increased)
    :param debug_log: Debug output during add_jitter CustomOp?
    :param test_intermediates: If given, all intermediates are written into
        this dict
    :return: L, P
    """
    kernel_mat = kernel(features, features)
#   Add jitter to noise_variance (if needed) in order to guarantee that
#   Cholesky factorization works
    sys_mat = AddJitterOp(
        flatten_and_concat(kernel_mat, noise_variance),
        initial_jitter_factor=NOISE_VARIANCE_LOWER_BOUND,
        debug_log='true' if debug_log else 'false')
    chol_fact = cholesky_factorization(sys_mat)
    centered_y = targets - anp.reshape(mean(features), (-1, 1))
    pred_mat = aspl.solve_triangular(chol_fact, centered_y, lower=True)
    # print('chol_fact', chol_fact)

    if test_intermediates is not None:
        assert isinstance(test_intermediates, dict)
        test_intermediates.update({
            'features': features,
            'targets': targets,
            'noise_variance': noise_variance,
            'kernel_mat': kernel_mat,
            'sys_mat': sys_mat,
            'chol_fact': chol_fact,
            'pred_mat': pred_mat,
            'centered_y': centered_y})
        test_intermediates.update(kernel.get_params())
        test_intermediates.update(mean.get_params())   
    
    return chol_fact, pred_mat


def predict_posterior_marginals(
        features, mean, kernel, chol_fact, pred_mat, test_features,
        test_intermediates=None):
    """
    Computes posterior means and variances for test_features.
    If pred_mat is a matrix, so will be posterior_means, but not
    posterior_variances. Reflects the fact that for GP regression and fixed
    hyperparameters, the posterior mean depends on the targets y, but the
    posterior covariance does not.

    :param features: Training inputs
    :param mean: Mean function
    :param kernel: Kernel function
    :param chol_fact: Part L of posterior state
    :param pred_mat: Part P of posterior state
    :param test_features: Test inputs
    :return: posterior_means, posterior_variances
    """
    k_tr_te = kernel(features, test_features)
    linv_k_tr_te = aspl.solve_triangular(chol_fact, k_tr_te, lower=True)
    posterior_means = anp.matmul(anp.transpose(linv_k_tr_te), pred_mat) + \
                      anp.reshape(mean(test_features), (-1, 1))
    posterior_variances = kernel.diagonal(test_features) - anp.sum(
        anp.square(linv_k_tr_te), axis=0)
    if test_intermediates is not None:
        assert isinstance(test_intermediates, dict)
        test_intermediates.update({
            'k_tr_te': k_tr_te,
            'linv_k_tr_te': linv_k_tr_te,
            'test_features': test_features,
            'pred_means': posterior_means,
            'pred_vars': anp.reshape(anp.maximum(
                posterior_variances, MIN_POSTERIOR_VARIANCE),
                (-1,))})    

    return posterior_means, anp.reshape(anp.maximum(
        posterior_variances, MIN_POSTERIOR_VARIANCE), (-1,))


def sample_posterior_marginals(
        features, mean, kernel, chol_fact, pred_mat, test_features,
        num_samples=1):
    """
    Draws num_sample samples from the product of marginals of the posterior
    over input points test_features. If pred_mat is a matrix with m columns,
    the samples returned have shape (n_test, m, num_samples).

    :param features: Training inputs
    :param mean: Mean function
    :param kernel: Kernel function
    :param chol_fact: Part L of posterior state
    :param pred_mat: Part P of posterior state
    :param test_features: Test inputs
    :param num_samples: Number of samples to draw
    :return: Samples, shape (n_test, num_samples) or (n_test, m, num_samples)
    """
    post_means, post_vars = predict_posterior_marginals(
        features, mean, kernel, chol_fact, pred_mat, test_features)
    post_means = anp.expand_dims(post_means, axis=-1)  # (n_test, m, 1)
    post_stds = anp.sqrt(anp.reshape(post_vars, (-1, 1, 1))) # (n_test, 1, 1)
    n01_vecs = [anp.random.normal(size=getval(post_means.shape))
                for _ in range(num_samples)]
    n01_mat = anp.concatenate(n01_vecs, axis=-1)
    samples = anp.multiply(n01_mat, post_stds) + post_means
    
    if samples.shape[1] == 1:
        n_test = getval(samples.shape)[0]
        samples = anp.reshape(samples, (n_test, -1))  # (n_test, num_samples)
        
    return samples


def sample_posterior_joint(
        features, mean, kernel, chol_fact, pred_mat, test_features,
        num_samples=1):
    """
    Draws num_sample samples from joint posterior distribution over inputs
    test_features. This is done by computing mean and covariance matrix of
    this posterior, and using the Cholesky decomposition of the latter. If
    pred_mat is a matrix with m columns, the samples returned have shape
    (n_test, m, num_samples).

    :param features: Training inputs
    :param mean: Mean function
    :param kernel: Kernel function
    :param chol_fact: Part L of posterior state
    :param pred_mat: Part P of posterior state
    :param test_features: Test inputs
    :param num_samples: Number of samples to draw
    :return: Samples, shape (n_test, num_samples) or (n_test, m, num_samples)
    """
    k_tr_te = kernel(features, test_features)
    linv_k_tr_te = aspl.solve_triangular(chol_fact, k_tr_te, lower=True)
    posterior_mean = anp.matmul(anp.transpose(linv_k_tr_te), pred_mat) + \
                     anp.reshape(mean(test_features), (-1, 1))
    posterior_cov = kernel(test_features, test_features) - anp.dot(
        anp.transpose(linv_k_tr_te), linv_k_tr_te)
    jitter_init = anp.ones((1,)) * (1e-5)
    sys_mat = AddJitterOp(
        flatten_and_concat(posterior_cov, jitter_init),
        initial_jitter_factor=NOISE_VARIANCE_LOWER_BOUND)
    lfact = cholesky_factorization(sys_mat)
    # Draw samples
    # posterior_mean.shape = (n_test, m), where m is number of cols of pred_mat
    # Reshape to (n_test, m, 1)
    n_test = getval(posterior_mean.shape)[0]
    posterior_mean = anp.expand_dims(posterior_mean, axis=-1)
    n01_vecs = [anp.random.normal(size=getval(posterior_mean.shape))
                for _ in range(num_samples)]
    n01_mat = anp.reshape(anp.concatenate(n01_vecs, axis=-1), (n_test, -1))
    samples = anp.reshape(
        anp.dot(lfact, n01_mat), (n_test, -1, num_samples))
    samples = samples + posterior_mean
    
    if samples.shape[1] == 1:
        samples = anp.reshape(samples, (n_test, -1))  # (n_test, num_samples)
        
    return samples


def _compute_lvec(features, chol_fact, kernel, feature):
    kvec = anp.reshape(kernel(features, feature), (-1, 1))
    return anp.reshape(aspl.solve_triangular(
        chol_fact, kvec, lower=True), (1, -1))


def cholesky_update(
        features, chol_fact, pred_mat, mean, kernel, noise_variance,
        feature, target, lvec=None):
    """
    Incremental update of posterior state (Cholesky factor, prediction
    matrix), given one datapoint (feature, target).

    Note: noise_variance is the initial value, before any jitter may have
    been added to compute chol_fact. Here, we add the minimum amount of
    jitter such that the new diagonal entry of the Cholesky factor is
    >= MIN_CHOLESKY_DIAGONAL_VALUE. This means that if cholesky_update is
    used several times, we in fact add a diagonal (but not spherical)
    jitter matrix.

    :param features: Shape (n, d)
    :param chol_fact: Shape (n, n)
    :param pred_mat: Shape (n, m)
    :param mean:
    :param kernel:
    :param noise_variance:
    :param feature: Shape (1, d)
    :param target: Shape (1, m)
    :param lvec: If given, this is the new column of the Cholesky factor
        except the diagonal entry. If not, this is computed here
    :return: chol_fact_new (n+1, n+1), pred_mat_new (n+1, m)
    """
    if lvec is None:
        lvec = _compute_lvec(features, chol_fact, kernel, feature)
    kscal = anp.reshape(kernel.diagonal(feature), (1,))
    noise_variance = anp.reshape(noise_variance, (1,))
    lsqscal = anp.maximum(
        kscal + noise_variance - anp.sum(anp.square(lvec)),
        MIN_CHOLESKY_DIAGONAL_VALUE ** 2)
    lscal = anp.reshape(anp.sqrt(lsqscal), (1, 1))
    mscal = anp.reshape(mean(feature), (1, 1))
    pvec = target - mscal
    pvec = anp.divide(pvec - anp.matmul(lvec, pred_mat), lscal)
    pred_mat_new = anp.concatenate([pred_mat, pvec], axis=0)
    zerovec = anp.zeros((getval(lvec.size), 1))
    chol_fact_new = anp.concatenate([
        anp.concatenate([chol_fact, lvec], axis=0),
        anp.concatenate([zerovec, lscal], axis=0)], axis=1)

    return chol_fact_new, pred_mat_new

# Specialized routine, used in IncrementalUpdateGPPosteriorState.
# The idea is to share the computation of lvec between sampling a new target
# value and incremental Cholesky update.
# If mean_impute_mask is given, it is a boolean vector of size m (number
# columns of pred_mat). Columns j of target, where mean_impute_ mask[j] is
# true, are set to the predictive mean (instead of being sampled).
def sample_and_cholesky_update(
        features, chol_fact, pred_mat, mean, kernel, noise_variance,
        feature, mean_impute_mask=None):
    # Draw sample target. Also, lvec is reused below
    lvec = _compute_lvec(features, chol_fact, kernel, feature)
    pred_mean = anp.dot(lvec, pred_mat) + anp.reshape(mean(feature), (1, 1))
    # Note: We do not add noise_variance to the predictive variance
    pred_std = anp.reshape(anp.sqrt(anp.maximum(
        kernel.diagonal(feature) - anp.sum(anp.square(lvec)),
        MIN_POSTERIOR_VARIANCE)), (1, 1))
    n01mat = anp.random.normal(size=getval(pred_mean.shape))
    if mean_impute_mask is not None:
        assert len(mean_impute_mask) == pred_mat.shape[1]
        n01mat[0, mean_impute_mask] = 0
    target = pred_mean + anp.multiply(n01mat, pred_std)
    chol_fact_new, pred_mat_new = cholesky_update(
        features, chol_fact, pred_mat, mean, kernel, noise_variance,
        feature, target, lvec=lvec)
    features_new = anp.concatenate([features, feature], axis=0)
    
    return chol_fact_new, pred_mat_new, features_new, target


def negative_log_marginal_likelihood(chol_fact, pred_mat):
    """
    The marginal likelihood is only computed if pred_mat has a single column
    (not for fantasy sample case).
    """
    assert pred_mat.ndim == 1 or pred_mat.shape[1] == 1,\
        "Multiple target vectors are not supported"
    sqnorm_predmat = anp.sum(anp.square(pred_mat))
    logdet_cholfact = 2.0 * anp.sum(anp.log(anp.abs(anp.diag(chol_fact))))
    n_samples = getval(pred_mat.size)

    return 0.5 * (sqnorm_predmat + n_samples * anp.log(2 * anp.pi) + logdet_cholfact)



import numpy as np
from typing import Dict, Optional

import logging
from .meanstd_acqfunc import MeanStdAcquisitionFunction, AcquisitionWithMultiModelCurrentBest, HeadWithGradient
from scipy.stats import norm
from ..tuning_algorithms.base_classes import OutputSurrogateModel, SurrogateModel
from ..utils.density import get_quantiles


logger = logging.getLogger(__name__)
MIN_COST = 1e-12   # For numerical stability when dividing EI / cost
MIN_STD_CONSTRAINT = 1e-12   # For numerical stability when computing the constraint probability in CEI


def _extract_active_and_secondary_metric(model_output_names, active_metric):
    """
    Returns the active metric and the secondary metric (such as the cost or constraint metric) from model_output_names.
    """

    assert len(model_output_names) == 2, f"The model should consist of exactly 2 outputs, " \
                                         f"while the current outputs are {model_output_names}"
    assert active_metric in model_output_names, f"{active_metric} is not a valid metric. " \
                                                f"The metric name must match one of the following metrics " \
                                                f"in the model output: {model_output_names}"
    if model_output_names[0] == active_metric:
        secondary_metric = model_output_names[1]
    else:
        secondary_metric = model_output_names[0]
    logger.info(f"There are two metrics in the output: {model_output_names}. "
                f"The metric to optimize was set to '{active_metric}'. "
                f"The secondary metric is assumed to be '{secondary_metric}'")
    return active_metric, secondary_metric


class EIAcquisitionFunction(MeanStdAcquisitionFunction):
    """
    Minus expected improvement acquisition function
    (minus because the convention is to always minimize acquisition functions)

    """
    def __init__(self, model: OutputSurrogateModel, active_metric: str = None,
                 jitter: float = 0.01):
        assert isinstance(model, SurrogateModel)
        super().__init__(model, active_metric)
        self.jitter = jitter

    def _head_needs_current_best(self) -> bool:
        return True

    def _compute_heads(self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
                       current_best: Optional[np.ndarray]) -> np.ndarray:
        means, stds = self._extract_active_metric_stats(output_to_mean_std)
        assert current_best is not None

        # phi, Phi is PDF and CDF of Gaussian
        phi, Phi, u = get_quantiles(self.jitter, current_best, means, stds)
        return (-stds) * (u * Phi + phi)

    def _compute_head_and_gradient(self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
                                   current_best: Optional[np.ndarray]) -> HeadWithGradient:
        mean, std = self._extract_active_metric_stats(output_to_mean_std)
        assert current_best is not None

        # phi, Phi is PDF and CDF of Gaussian
        phi, Phi, u = get_quantiles(self.jitter, current_best, mean, std)
        f_acqu = std * (u * Phi + phi)
        return HeadWithGradient(
            hvals=-f_acqu,
            dh_dmean={self.active_metric: Phi},
            dh_dstd={self.active_metric: -phi})


class LCBAcquisitionFunction(MeanStdAcquisitionFunction):
    """
    Lower confidence bound (LCB) acquisition function:

        h(mean, std) = mean - kappa * std

    """
    def __init__(self, model: OutputSurrogateModel, kappa: float, active_metric: str = None):
        super().__init__(model, active_metric)
        assert kappa > 0, 'kappa must be positive'
        self.kappa = kappa

    def _head_needs_current_best(self) -> bool:
        return False

    def _compute_heads(self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
                       current_best: Optional[np.ndarray]) -> np.ndarray:
        means, stds, _ = self._extract_active_metric_stats(output_to_mean_std)
        return means - stds * self.kappa

    def _compute_head_and_gradient(self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
                                   current_best: Optional[np.ndarray]) -> HeadWithGradient:
        mean, std = self._extract_active_metric_stats(output_to_mean_std)
        ones_like_mean = np.ones_like(mean)
        ones_like_std = np.ones_like(std)
        return HeadWithGradient(
            hvals=mean - std * self.kappa,
            dh_dmean={self.active_metric: ones_like_mean},
            dh_dstd={self.active_metric: (-self.kappa) * ones_like_std})


class EIpuAcquisitionFunction(MeanStdAcquisitionFunction):
    """
    Minus cost-aware expected improvement acquisition function.
    (minus because the convention is to always minimize the acquisition function)

    This is defined as EIpu(x) = EI(x) / cost(x), where cost(x) is the predictive mean of the cost model at x.

    Note: two metrics are expected in the model output: the main objective and the cost.
    The main objective needs to be indicated as active_metric when initializing EIpuAcquisitionFunction.
    The cost is automatically assumed to be the other metric.

    """
    def __init__(self, model: OutputSurrogateModel, active_metric: str, jitter: float = 0.01):
        super(EIpuAcquisitionFunction, self).__init__(model, active_metric)
        self.jitter = jitter
        self.active_metric, self.cost_metric = _extract_active_and_secondary_metric(
            self.model_output_names, active_metric)

    def _head_needs_current_best(self) -> bool:
        return True

    def _compute_heads(
            self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
            current_best: Optional[np.ndarray]) -> np.ndarray:
        """
        Returns minus the cost-aware expected improvement.
        """
        means, stds = self._extract_active_metric_stats(output_to_mean_std)
        assert current_best is not None
        pred_costs = self._extract_positive_cost(output_to_mean_std)

        # phi, Phi is PDF and CDF of Gaussian
        phi, Phi, u = get_quantiles(self.jitter, current_best, means, stds)
        f_ei = stds * (u * Phi + phi)
        f_acqu = f_ei / pred_costs
        return - f_acqu

    def _compute_head_and_gradient(
            self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
            current_best: Optional[np.ndarray]) -> HeadWithGradient:
        """
        Returns minus cost-aware expected improvement and, for each output model, the gradients
        with respect to the mean and standard deviation of that model.
        """
        mean, std = self._extract_active_metric_stats(output_to_mean_std)
        assert current_best is not None
        pred_cost = self._extract_positive_cost(output_to_mean_std)

        # phi, Phi is PDF and CDF of Gaussian
        phi, Phi, u = get_quantiles(self.jitter, current_best, mean, std)
        f_ei = std * (u * Phi + phi)
        f_acqu = f_ei / pred_cost

        dh_dmean_active = Phi / pred_cost
        dh_dstd_active = - phi / pred_cost
        dh_dmean_cost = f_ei / (pred_cost ** 2)  # We flip the sign twice: once because of the derivative of 1 / x
        # and once because the head is actually - f_ei
        dh_dstd_cost = np.zeros_like(dh_dstd_active)   # EIpu does not depend on the standard deviation of cost

        return HeadWithGradient(
            hvals=-f_acqu,
            dh_dmean={self.active_metric: dh_dmean_active, self.cost_metric: dh_dmean_cost},
            dh_dstd={self.active_metric: dh_dstd_active, self.cost_metric: dh_dstd_cost}
        )

    def _extract_positive_cost(self, output_to_mean_std_best):
        pred_cost = output_to_mean_std_best[self.cost_metric]['mean']
        if any(pred_cost) < 0.0:
            logger.warning(f'The model for {self.cost_metric} predicted some negative cost. '
                           f'Capping the minimum cost at {MIN_COST}.')
        pred_cost = np.maximum(pred_cost, MIN_COST)  # ensure that the predicted cost/run-time is positive
        return pred_cost


class CEIAcquisitionFunction(AcquisitionWithMultiModelCurrentBest):
    """
    Minus constrained expected improvement acquisition function.
    (minus because the convention is to always minimize the acquisition function)

    This is defined as CEI(x) = EI(x) * P(c(x) <= 0), where EI is the standard expected improvement with respect
    to the current *feasible best*, and P(c(x) <= 0) is the probability that the hyperparameter
    configuration x satisfies the constraint modeled by c(x).

    If there are no feasible hyperparameters yet, the current feasible best is undefined. Thus, CEI is
    reduced to the P(c(x) <= 0) term until a feasible configuration is found.

    Two metrics are expected in the model output: the main objective and the constraint metric.
    The main objective needs to be indicated as active_metric when initializing CEIAcquisitionFunction.
    The constraint is automatically assumed to be the other metric.

    References on CEI:
    Gardner et al., Bayesian Optimization with Inequality Constraints. In ICML, 2014.
    Gelbart et al., Bayesian Optimization with Unknown Constraints. In UAI, 2014.

    """
    def __init__(self, model: OutputSurrogateModel, active_metric: str, jitter: float = 0.01):
        super(CEIAcquisitionFunction, self).__init__(model, active_metric)
        self.jitter = jitter
        self._feasible_best_list = None
        self.active_metric, self.constraint_metric = _extract_active_and_secondary_metric(
            self.model_output_names, active_metric)

    def _head_needs_current_best(self) -> bool:
        return True

    def _compute_heads(self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
                       current_best: Optional[np.ndarray]) -> np.ndarray:
        """
        Returns minus the constrained expected improvement (- CEI).
        """
        assert current_best is not None
        means, stds = self._extract_active_metric_stats(output_to_mean_std)
        means_constr = output_to_mean_std[self.constraint_metric]['mean']
        stds_constr = output_to_mean_std[self.constraint_metric]['std']
        # Compute the probability of satisfying the constraint P(c(x) <= 0)
        constr_probs = norm.cdf(- means_constr / (stds_constr + MIN_STD_CONSTRAINT))
        # If for some fantasies there are not feasible candidates, there is also no current_best (i.e., a nan).
        # The acquisition function is replaced by only the P(c(x) <= 0) term when no feasible best exist.
        feas_idx = ~np.isnan(current_best).flatten()
        num_fantasies = current_best.size
        means = means.reshape((-1, num_fantasies))
        stds = stds.reshape((-1, 1))
        current_best = current_best.reshape((1, num_fantasies))
        constr_probs = constr_probs.reshape((-1, num_fantasies))

        # phi, Phi is PDF and CDF of Gaussian
        phi, Phi, u = get_quantiles(self.jitter, current_best, means, stds)
        f_ei = stds * (u * Phi + phi)
        # CEI(x) = EI(x) * P(c(x) <= 0) if feasible best exists, CEI(x) = P(c(x) <= 0) otherwise
        f_acqu = np.where(feas_idx, f_ei * constr_probs, constr_probs)
        return - f_acqu

    def _compute_head_and_gradient(self, output_to_mean_std: Dict[str, Dict[str, np.ndarray]],
                                   current_best: Optional[np.ndarray]) -> HeadWithGradient:
        """
        Returns minus cost-aware expected improvement (- CEI) and, for each output model, the gradients
        with respect to the mean and standard deviation of that model.
        """
        assert current_best is not None
        mean, std = self._extract_active_metric_stats(output_to_mean_std)
        mean_constr = output_to_mean_std[self.constraint_metric]['mean']
        std_constr = output_to_mean_std[self.constraint_metric]['std']
        # Compute the probability of satisfying the constraint P(c(x) <= 0)
        std_constr = std_constr + MIN_STD_CONSTRAINT
        z = - mean_constr / std_constr
        constr_prob = norm.cdf(z)
        # Useful variables for computing the head gradients
        mean_over_squared_std_constr = mean_constr / std_constr ** 2
        inverse_std_constr = 1. / std_constr
        phi_constr = norm.pdf(z)

        # If for some fantasies there are not feasible candidates, there is also no current_best (i.e., a nan).
        # The acquisition function is replaced by only the P(c(x) <= 0) term when no feasible best exist.
        feas_idx = ~np.isnan(current_best)
        phi, Phi, u = get_quantiles(self.jitter, current_best, mean, std)  # phi, Phi is PDF and CDF of Gaussian
        f_ei = std * (u * Phi + phi)
        f_acqu = np.where(feas_idx, f_ei * constr_prob, constr_prob)  # CEI(x) = EI(x) * P(c(x) <= 0) if feasible best
        # exists, CEI(x) = P(c(x) <= 0) otherwise
        dh_dmean_constraint_feas = f_ei * inverse_std_constr * phi_constr
        dh_dstd_constraint_feas = - f_ei * mean_over_squared_std_constr * phi_constr
        dh_dmean_active_feas = Phi * constr_prob
        dh_dstd_active_feas = - phi * constr_prob
        dh_dmean_constraint_infeas = inverse_std_constr * phi_constr
        dh_dstd_constraint_infeas = - mean_over_squared_std_constr * phi_constr
        dh_dmean_active_infeas = np.zeros_like(phi_constr)
        dh_dstd_active_infeas = np.zeros_like(phi_constr)
        dh_dmean_active = np.where(feas_idx, dh_dmean_active_feas, dh_dmean_active_infeas)
        dh_dstd_active = np.where(feas_idx, dh_dstd_active_feas, dh_dstd_active_infeas)
        dh_dmean_constraint = np.where(feas_idx, dh_dmean_constraint_feas, dh_dmean_constraint_infeas)
        dh_dstd_constraint = np.where(feas_idx, dh_dstd_constraint_feas, dh_dstd_constraint_infeas)
        return HeadWithGradient(
            hvals=-f_acqu,
            dh_dmean={self.active_metric: dh_dmean_active, self.constraint_metric: dh_dmean_constraint},
            dh_dstd={self.active_metric: dh_dstd_active, self.constraint_metric: dh_dstd_constraint}
        )

    def _get_current_best_for_active_metric(self):
        """
        Returns a list of current best, one per MCMC sample (if head requires the current best).

        Filters out the infeasible candidates when computing the current best. This is needed as the constrained
        expected improvement uses the current best over the *feasible* hyperparameter
        configurations (i.e., satisfying the constraint) and not over all hyperparameter configurations.

        The assumption is that a hyperparameter configuration is feasible iff the predictive mean of the
        constraint model is <= 0 at that configuration.
        """
        if self._head_needs_current_best():
            if self._feasible_best_list is not None:
                feasible_best_list = self._feasible_best_list
            else:
                models_to_means = self._map_models_to_candidate_predictive_means()
                assert len(models_to_means) == 2, 'The model should consist of exactly 2 outputs.'
                all_means_active = models_to_means[self.active_metric]
                all_means_constraint = models_to_means[self.constraint_metric]
                assert len(all_means_constraint) == len(all_means_constraint), \
                    'All models must have the same number of MCMC samples.'
                feasible_best_list = []
                # Loop over MCMC samples (if any)
                for means_active, means_constraint in zip(all_means_active, all_means_constraint):
                    assert means_active.shape == means_constraint.shape, \
                        'The predictive means from each model must have the same shape.'
                    # Remove all infeasible candidates (i.e., where means_constraint is >= 0)
                    means_active[means_constraint >= 0] = np.nan
                    # Compute the current *feasible* best (separately for every fantasy)
                    min_across_observations = np.nanmin(means_active, axis=0)
                    feasible_best_list.append(min_across_observations)
                    self._feasible_best_list = feasible_best_list
        else:
            feasible_best_list = [None] * self.num_mcmc_samples
        return feasible_best_list



"""ssh helper for starting remote dask scheduler"""
from __future__ import print_function, division, absolute_import

import os
import socket
import sys
import time
import traceback

try:
    from queue import Queue
except ImportError:  # Python 2.7 fix
    from Queue import Queue
import logging

from threading import Thread
from toolz import merge

logger = logging.getLogger(__name__)

class bcolors:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

def async_ssh(cmd_dict):
    import paramiko
    from paramiko.buffered_pipe import PipeTimeout
    from paramiko.ssh_exception import SSHException, PasswordRequiredException

    ssh = paramiko.SSHClient()
    ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())

    retries = 0
    while True:  # Be robust to transient SSH failures.
        try:
            # Set paramiko logging to WARN or higher to squelch INFO messages.
            logging.getLogger("paramiko").setLevel(logging.WARN)

            ssh.connect(
                hostname=cmd_dict["address"],
                username=cmd_dict["ssh_username"],
                port=cmd_dict["ssh_port"],
                key_filename=cmd_dict["ssh_private_key"],
                compress=True,
                timeout=20,
                banner_timeout=20,
            )  # Helps prevent timeouts when many concurrent ssh connections are opened.
            # Connection successful, break out of while loop
            break

        except (SSHException, PasswordRequiredException) as e:

            print(
                "[ dask-ssh ] : "
                + bcolors.FAIL
                + "SSH connection error when connecting to {addr}:{port}"
                "to run '{cmd}'".format(
                    addr=cmd_dict["address"],
                    port=cmd_dict["ssh_port"],
                    cmd=cmd_dict["cmd"],
                )
                + bcolors.ENDC
            )

            print(
                bcolors.FAIL
                + "               SSH reported this exception: "
                + str(e)
                + bcolors.ENDC
            )

            # Print an exception traceback
            traceback.print_exc()

            # Transient SSH errors can occur when many SSH connections are
            # simultaneously opened to the same server. This makes a few
            # attempts to retry.
            retries += 1
            if retries >= 3:
                print(
                    "[ dask-ssh ] : "
                    + bcolors.FAIL
                    + "SSH connection failed after 3 retries. Exiting."
                    + bcolors.ENDC
                )

                # Connection failed after multiple attempts.  Terminate this thread.
                os._exit(1)

            # Wait a moment before retrying
            print(
                "               "
                + bcolors.FAIL
                + "Retrying... (attempt {n}/{total})".format(n=retries, total=3)
                + bcolors.ENDC
            )

            time.sleep(1)

    # Execute the command, and grab file handles for stdout and stderr. Note
    # that we run the command using the user's default shell, but force it to
    # run in an interactive login shell, which hopefully ensures that all of the
    # user's normal environment variables (via the dot files) have been loaded
    # before the command is run. This should help to ensure that important
    # aspects of the environment like PATH and PYTHONPATH are configured.

    print("[ {label} ] : {cmd}".format(label=cmd_dict["label"], cmd=cmd_dict["cmd"]))
    stdin, stdout, stderr = ssh.exec_command(
        "$SHELL -i -c '" + cmd_dict["cmd"] + "'", get_pty=True
    )

    # Set up channel timeout (which we rely on below to make readline() non-blocking)
    channel = stdout.channel
    channel.settimeout(0.1)

    def read_from_stdout():
        """
        Read stdout stream, time out if necessary.
        """
        try:
            line = stdout.readline()
            while len(line) > 0:  # Loops until a timeout exception occurs
                line = line.rstrip()
                logger.debug("stdout from ssh channel: %s", line)
                cmd_dict["output_queue"].put(
                    "[ {label} ] : {output}".format(
                        label=cmd_dict["label"], output=line
                    )
                )
                line = stdout.readline()
        except (PipeTimeout, socket.timeout):
            pass

    def read_from_stderr():
        """
        Read stderr stream, time out if necessary.
        """
        try:
            line = stderr.readline()
            while len(line) > 0:
                line = line.rstrip()
                logger.debug("stderr from ssh channel: %s", line)
                cmd_dict["output_queue"].put(
                    "[ {label} ] : ".format(label=cmd_dict["label"])
                    + bcolors.FAIL
                    + "{output}".format(output=line)
                    + bcolors.ENDC
                )
                line = stderr.readline()
        except (PipeTimeout, socket.timeout):
            pass

    def communicate():
        """
        Communicate a little bit, without blocking too long.
        Return True if the command ended.
        """
        read_from_stdout()
        read_from_stderr()

        # Check to see if the process has exited. If it has, we let this thread
        # terminate.
        if channel.exit_status_ready():
            exit_status = channel.recv_exit_status()
            cmd_dict["output_queue"].put(
                "[ {label} ] : ".format(label=cmd_dict["label"])
                + bcolors.FAIL
                + "remote process exited with exit status "
                + str(exit_status)
                + bcolors.ENDC
            )
            return True

    # Get transport to current SSH client
    transport = ssh.get_transport()

    # Wait for a message on the input_queue. Any message received signals this
    # thread to shut itself down.
    while cmd_dict["input_queue"].empty():
        # Kill some time so that this thread does not hog the CPU.
        time.sleep(1.0)
        # Send noise down the pipe to keep connection active
        transport.send_ignore()
        if communicate():
            break

    # Ctrl-C the executing command and wait a bit for command to end cleanly
    start = time.time()
    while time.time() < start + 5.0:
        try:
            channel.send(b"\x03")  # Ctrl-C
        except Exception:
            break
        if communicate():
            break
        time.sleep(1.0)

    # Shutdown the channel, and close the SSH connection
    channel.close()
    ssh.close()


def start_scheduler(addr, port, ssh_username, ssh_port,
                    ssh_private_key, remote_python=None):
    cmd = "dask-scheduler --port {port}".format(
        #python=remote_python or sys.executable,
        port=port
    )

    # Format output labels we can prepend to each line of output, and create
    # a 'status' key to keep track of jobs that terminate prematurely.
    label = (
        bcolors.BOLD
        + "scheduler {addr}:{port}".format(addr=addr, port=port)
        + bcolors.ENDC
    )

    # Create a command dictionary, which contains everything we need to run and
    # interact with this command.
    input_queue = Queue()
    output_queue = Queue()
    cmd_dict = {
        "cmd": cmd,
        "label": label,
        "address": addr,
        "port": port,
        "input_queue": input_queue,
        "output_queue": output_queue,
        "ssh_username": ssh_username,
        "ssh_port": ssh_port,
        "ssh_private_key": ssh_private_key,
    }

    # Start the thread
    thread = Thread(target=async_ssh, args=[cmd_dict])
    thread.daemon = True
    thread.start()

    return merge(cmd_dict, {"thread": thread})

def start_worker(scheduler_addr, scheduler_port, worker_addr,
    ssh_username, ssh_port, ssh_private_key,
    remote_python=None):

    cmd = (
        "dask-worker "
        "{scheduler_addr}:{scheduler_port} "
        "--no-nanny "
    )

    #if not nohost:
    cmd += " --host {worker_addr}"

    cmd = cmd.format(
        #python=remote_python or sys.executable,
        scheduler_addr=scheduler_addr,
        scheduler_port=scheduler_port,
        worker_addr=worker_addr,
    )

    label = "worker {addr}".format(addr=worker_addr)

    # Create a command dictionary, which contains everything we need to run and
    # interact with this command.
    input_queue = Queue()
    output_queue = Queue()
    cmd_dict = {
        "cmd": cmd,
        "label": label,
        "address": worker_addr,
        "input_queue": input_queue,
        "output_queue": output_queue,
        "ssh_username": ssh_username,
        "ssh_port": ssh_port,
        "ssh_private_key": ssh_private_key,
    }

    # Start the thread
    thread = Thread(target=async_ssh, args=[cmd_dict])
    thread.daemon = True
    thread.start()

    return merge(cmd_dict, {"thread": thread})



from typing import NamedTuple, List, Dict
import time
import logging
import numpy as np

logger = logging.getLogger(__name__)


class ProfilingBlock(NamedTuple):
    meta: dict
    time_stamp: float
    durations: Dict[str, List[float]]


class SimpleProfiler(object):
    """
    Useful to profile time of recurring computations, for example
    `get_config` calls in searchers.

    Measurements are divided into blocks. A block is started by `begin_block`.
    Each block stores meta data, a time stamp when `begin_block` was called
    (relative to the time stamp for the first block, which is 0), and a dict of
    lists of durations, whose keys are tags. A tag corresponds to a range of
    code to be profiled. It may be executed many times within a block,
    therefore lists of durations.

    Tags can have multiple levels of prefixes, corresponding to brackets.

    """

    def __init__(self):
        self.records = list()
        self.start_time = dict()
        self.time_stamp_first_block = None
        self.meta_keys = None
        self.prefix = ''

    def begin_block(self, meta: dict):
        assert not self.start_time, \
            "Timers for these tags still running:\n{}".format(
                self.start_time.keys())
        meta_keys = tuple(sorted(meta.keys()))
        if self.time_stamp_first_block is None:
            self.meta_keys = meta_keys
            self.time_stamp_first_block = time.time()
        else:
            assert meta_keys == self.meta_keys, \
                "meta.keys() = {}, but must be the same as for all previous meta dicts ({})".format(
                    meta_keys, self.meta_keys)
        time_stamp = time.time() - self.time_stamp_first_block
        new_block = ProfilingBlock(
            meta=meta.copy(),
            time_stamp=time_stamp,
            durations=dict())
        self.records.append(new_block)
        self.prefix = ''

    def push_prefix(self, prefix: str):
        assert '_' not in prefix, "Prefix must not contain '_'"
        self.prefix += (prefix + '_')

    def pop_prefix(self):
        lpref = len(self.prefix)
        assert lpref > 0, "Prefix is empty"
        pos = self.prefix.rfind('_', 0, lpref - 1)
        if pos != -1:
            self.prefix = self.prefix[:(pos + 1)]
        else:
            self.prefix = ''

    def start(self, tag: str):
        assert self.records, \
            "No block has been started yet (use 'begin_block')"
        tag = self.prefix + tag
        assert tag not in self.start_time, \
            "Timer for '{}' already running".format(tag)
        self.start_time[tag] = time.time()

    def stop(self, tag: str):
        assert self.records, \
            "No block has been started yet (use 'begin_block')"
        tag = self.prefix + tag
        assert tag in self.start_time, \
            "Timer for '{}' does not exist".format(tag)
        duration = time.time() - self.start_time[tag]
        block = self.records[-1]
        if tag in block.durations:
            block.durations[tag].append(duration)
        else:
            block.durations[tag] = [duration]
        del self.start_time[tag]

    def clear(self):
        remaining_tags = list(self.start_time.keys())
        if remaining_tags:
            logger.warning("Timers for these tags not stopped (will be removed):\n{}".format(
                remaining_tags))
        self.start_time = dict()

    def records_as_dict(self) -> dict:
        """
        Return records as a dict of lists, can be converted into Pandas
        data-frame by:

            pandas.DataFrame.fromDict(...)

        Each entry corresponds to a column.
        """
        if len(self.records) == 0:
            return dict()
        # For each tag, we emit the following columns: tag_num, tag_mean,
        # tag_std
        data = {'time_stamp': []}
        union_tags = self._union_of_tags()
        suffices = ('_num', '_mean', '_std')
        for tag in union_tags:
            for suffix in suffices:
                data[tag + suffix] = []
        for k in self.meta_keys:
            data[k] = []
        # fill the columns row by row
        for block in self.records:
            data['time_stamp'].append(block.time_stamp)
            for k, v in block.meta.items():
                data[k].append(v)
            for tag in union_tags:
                for suffix in suffices:
                    data[tag + suffix].append(0)
            for tag, durations in block.durations.items():
                data[tag + '_num'][-1] = len(durations)
                data[tag + '_mean'][-1] = np.mean(durations)
                data[tag + '_std'][-1] = np.std(durations)
        return data

    def _union_of_tags(self):
        union_tags = set()
        for block in self.records:
            union_tags.update(block.durations.keys())
        return union_tags
